/* Matrix multiply kernel */
__kernel void
mat_mul_simple(__global float* matA,
		__global float* matB,
		__global float* matC,
		int size)
{
	int gid_0 = get_global_id(0);
	int gid_1 = get_global_id(1);
	
	matC [gid_1 * size + gid_0] = 0.f;
	for(int k = 0; k < size; k++) {
		matC [gid_1 * size + gid_0] += matA [gid_0 * size + k] * matB [k * size + gid_1];
	}
}

/* Matrix multiply kernel */
__kernel void
mat_mul(__global float* matA,
		__global float* matB,
		__global float* matC,
		int size)
{
	int gid_0 = get_global_id(0);
	int gid_1 = get_global_id(1);
	
	matC [gid_1 * size + gid_0] = 0.f;
	float4 vecA;
	float4 vecB;
	float4 vecC;
	
	vecC = 0.f;
	
	for(int k = 0; k < size; k+=4) {
		vecA.s0 = matA [gid_0 * size + k + 0];
		vecA.s1 = matA [gid_0 * size + k + 1];
		vecA.s2 = matA [gid_0 * size + k + 2];
		vecA.s3 = matA [gid_0 * size + k + 3];
		
		vecB.s0 = matB [k * size + gid_1];
		vecB.s1 = matB [k * size + gid_1];
		vecB.s2 = matB [k * size + gid_1];
		vecB.s3 = matB [k * size + gid_1];
		
		vecC += vecA * vecB;
	}
	
	matC [gid_1 * size + gid_0] = vecC.s0 + vecC.s1 + vecC.s2 + vecC.s3;
}
